class Binary:
    def __init__(self) -> None:
        pass

class Ternary:
    def __init__(self) -> None:
        pass

class Ternary3D:
    def __init__(self) -> None:
        pass

class Quaternery:
    def __init__(self) -> None:
        pass